from units import *
import math

class Link:

   def __init__(self, timestep, freeFlowSpeed, backwardWaveSpeed, jamDensity, length, capacity):
      """
      Sets up a link, adjusting units in the following way:
         INPUTS:
         freeFlowSpeed and backwardWaveSpeed arguments must be provided in mi/hr
         jamDensity must be in veh/mi
         length must be provided in feet
         timestep must be in seconds
         capacity argument must be provided in veh/hr
         
         ATTRIBUTES:
         freeFlowSpeed and backwardWaveSpeed will be in *ft/timestep*
         jamDensity will be in *veh/ft*
         length will be in *ft*
         freeFlowTime and backwardWaveTime will be integer numbers of *timesteps*
         capacity will be in vehicles per *timestep*         
      """ 
      self.capacity = capacity / HOURS * timestep
      self.freeFlowSpeed = freeFlowSpeed * MILES / HOURS * timestep
      self.backwardWaveSpeed = backwardWaveSpeed * MILES / HOURS * timestep
      self.jamDensity = jamDensity / MILES
      self.length = length * FEET
      self.maxVehicles = self.length * self.jamDensity
      
      # freeFlowTime and backwardWaveTime are in units of TIMESTEPs
      self.freeFlowTime = math.ceil(self.length/self.freeFlowSpeed) 
      self.backwardWaveTime = math.ceil(self.length/self.backwardWaveSpeed)
      
      # initialize dictionaries and counts
      self.upstreamPathCount = list()
      self.downstreamPathCount = list()
      self.upstreamPathCount.append(dict())
      self.downstreamPathCount.append(dict())

   def calculateSendingFlow(self, t):
      """
      This method is implemented differently for each link model, and must be overridden in the derived classes.
      """
      pass

   def calculateReceivingFlow(self, t):
      """
      This method is implemented differently for each link model, and must be overridden in the derived classes.
      """
   
      pass

   def linkUpdate(self, t):
      """
      linkUpdate performs any internal calculations needed for a link at time t, and returns a tuple (S,R) with the sending and receiving flows
      By default, it only returns this tuple; if more is needed you need to override this method.
      """
      return (self.calculateSendingFlow(t), self.calculateReceivingFlow(t))

   def upstreamCount(self, t):
      """
      upstreamCount calculates the total number of vehicles that have entered the link, up through time t.
      You should only call this method for an integer value of t.
      """
      t = math.floor(t)
      if t < 0:
         return 0
      count = 0
      for path in self.upstreamPathCount[t].keys():
         count += self.upstreamPathCount[t][path]
      return count

   def downstreamCount(self, t):
      """
      downstreamCount calculates the total number of vehicles that have exited the link, up through time t.
      You should only call this method for an integer value of t.
      """
      t = math.floor(t)   
      if t < 0:
         return 0
      count = 0
      for path in self.downstreamPathCount[t].keys():
         count += self.downstreamPathCount[t][path]
      return count

   def vehiclesOnLink(self, t):
      """
      vehiclesOnLink calculates the total number of vehicles on the link at time t.
      You should only call this method for an integer value of t.
      """
      return self.upstreamCount(t) - self.downstreamCount(t)

   def flowIn(self, pathFlows):
      """
      Handles vehicles entering a link; you do not need to understand the details of this implementation.
      Adds flow to the upstream end of a link; based on pathFlows.  Extends upstream array.
      """
      self.upstreamPathCount.append(dict())
      # First copy previous flows
      for path in self.upstreamPathCount[-2].keys():
         self.upstreamPathCount[-1][path] = self.upstreamPathCount[-2][path]

      # Then add new inflows
      for path in pathFlows:
         self.upstreamPathCount[-1][path] = pathFlows[path] + self.upstreamPathCount[-1].setdefault(path, 0)

   def flowOut(self, pathFlows):
      """
      Handles vehicles leaving a link; you do not need to understand the details of this implementation.
      Removes flow from the downstream end of a link; based on pathFlows.  Extends downstream array
      """
      self.downstreamPathCount.append(dict())
      # First copy previous flows
      for path in self.upstreamPathCount[-2].keys():
         self.downstreamPathCount[-1][path] = self.downstreamPathCount[-2][path]

      # Then add new outflows
      for path in pathFlows:
         self.downstreamPathCount[-1][path] = pathFlows[path] + self.downstreamPathCount[-1].setdefault(path, 0)


