import sys
import traceback

import link
import linkModel
import utils
from units import *
   
IS_MISSING = -1

def approxEqual(value, target, tolerance):
   if (abs(target) <= tolerance ): return abs(value) <= tolerance
   return abs(float(value) / target) - 1 <= tolerance
   
def check(name, t, value, target, tolerance):
   if approxEqual(value, target, tolerance):
      return True
   else:
      print("\nWrong %s at time %d: your value %f, correct value %f"
               % (name, t, value, target))
      return False

def PQ(testFileName):
  
   print("Running point queue test: " + str(testFileName) + "...", end='')
   inflow = list()
   targetSendingFlow = list()
   targetReceivingFlow = list()
   targetUpstreamCount = list()
   targetDownstreamCount = list()
   
   try:
      with open(testFileName, "r") as testFile:
         # Read test information
         try:
            fileLines = testFile.read().splitlines()
            pointsPossible = IS_MISSING
            timeHorizon = IS_MISSING
            for line in fileLines:
               # Ignore comments and blank lines
               if len(line.strip()) == 0 or line[0] == '#':
                  continue
                   
               # Set points possible
               if pointsPossible == IS_MISSING:
                  pointsPossible = int(line)
                  continue
                  
               # Set problem parameters
               if timeHorizon == IS_MISSING:
                  inputs = line.split(",")
                  timeHorizon = int(inputs[0])
                  freeFlowTime = int(inputs[1])
                  downstreamCapacity = float(inputs[2])
                  upstreamCapacity = float(inputs[3])
                  continue
                  
               # Read problem data
               timestepData = line.split(",")
               inflow.append(float(timestepData[0]))
               targetUpstreamCount.append(float(timestepData[1]))
               targetDownstreamCount.append(float(timestepData[2]))
               targetReceivingFlow.append(float(timestepData[3]))
               targetSendingFlow.append(float(timestepData[4]))               
                  
         except:
            print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
            traceback.print_exc(file=sys.stdout)
            return 0, 0
            
         # Now run the actual test
         try:
            testLink = linkModel.PointQueueLink(1 * HOURS, 1, 1, 1, freeFlowTime * MILES, downstreamCapacity, upstreamCapacity)
            enteringFlow = dict()
            leavingFlow = dict()
            for t in range(timeHorizon):
               upstreamCount, downstreamCount = testLink.upstreamCount(t), testLink.downstreamCount(t)
               sendingFlow, receivingFlow = testLink.linkUpdate(t)

               enteringFlow['link'] = min(inflow[t], receivingFlow)
               leavingFlow['link'] = sendingFlow
              
               testLink.flowIn(enteringFlow)
               testLink.flowOut(leavingFlow)
              
               correct = True
               correct = correct and check("sending flow", t, sendingFlow, targetSendingFlow[t], 0.01)
               correct = correct and check("receiving flow", t, receivingFlow, targetReceivingFlow[t], 0.01)            
               correct = correct and check("upstream count", t, upstreamCount, targetUpstreamCount[t], 0.01)            
               correct = correct and check("downstream count", t, downstreamCount, targetDownstreamCount[t], 0.01)            
               if correct == False:               
                  print("...fail")
                  return 0, pointsPossible
         except utils.NotYetAttemptedException:
            print("...not yet attempted")
            return 0, pointsPossible
         except:
            print("\nException raised, attempting to continue:")
            traceback.print_exc(file=sys.stdout)                     
            print("\n...fail")
            return 0, pointsPossible
            
         print("...pass")
         return pointsPossible, pointsPossible

         
   except IOError:
      print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
      traceback.print_exc(file=sys.stdout) 
      return 0, 0
   

def SQ(testFileName):
  
   print("Running spatial queue test: " + str(testFileName) + "...", end='')
   inflow = list()
   targetSendingFlow = list()
   targetReceivingFlow = list()
   targetUpstreamCount = list()
   targetDownstreamCount = list()
   
   try:
      with open(testFileName, "r") as testFile:
         # Read test information
         try:
            fileLines = testFile.read().splitlines()
            pointsPossible = IS_MISSING
            timeHorizon = IS_MISSING
            for line in fileLines:
               # Ignore comments and blank lines
               if len(line.strip()) == 0 or line[0] == '#':
                  continue
                   
               # Set points possible
               if pointsPossible == IS_MISSING:
                  pointsPossible = int(line)
                  continue
                  
               # Set problem parameters
               if timeHorizon == IS_MISSING:
                  inputs = line.split(",")
                  timeHorizon = int(inputs[0])
                  freeFlowTime = int(inputs[1])
                  downstreamCapacity = float(inputs[2])
                  upstreamCapacity = float(inputs[3])
                  jamDensity = float(inputs[4])
                  continue
                  
               # Read problem data
               timestepData = line.split(",")
               inflow.append(float(timestepData[0]))
               targetUpstreamCount.append(float(timestepData[1]))
               targetDownstreamCount.append(float(timestepData[2]))
               targetReceivingFlow.append(float(timestepData[3]))
               targetSendingFlow.append(float(timestepData[4]))               
                  
         except:
            print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
            traceback.print_exc(file=sys.stdout)
            return 0, 0
            
         # Now run the actual test
         try:
            testLink = linkModel.SpatialQueueLink(1 * HOURS, 1, 1, jamDensity, freeFlowTime * MILES, downstreamCapacity, upstreamCapacity)
            enteringFlow = dict()
            leavingFlow = dict()
            for t in range(timeHorizon):
               upstreamCount, downstreamCount = testLink.upstreamCount(t), testLink.downstreamCount(t)
               sendingFlow, receivingFlow = testLink.linkUpdate(t)

               enteringFlow['link'] = min(inflow[t], receivingFlow)
               leavingFlow['link'] = sendingFlow
              
               testLink.flowIn(enteringFlow)
               testLink.flowOut(leavingFlow)
              
               correct = True
               correct = correct and check("sending flow", t, sendingFlow, targetSendingFlow[t], 0.01)
               correct = correct and check("receiving flow", t, receivingFlow, targetReceivingFlow[t], 0.01)            
               correct = correct and check("upstream count", t, upstreamCount, targetUpstreamCount[t], 0.01)            
               correct = correct and check("downstream count", t, downstreamCount, targetDownstreamCount[t], 0.01)            
               if correct == False:               
                  print("\n...fail")
                  return 0, pointsPossible
         except utils.NotYetAttemptedException:
            print("...not yet attempted")
            return 0, pointsPossible
         except:
            print("\nException raised, attempting to continue:")
            traceback.print_exc(file=sys.stdout)                     
            print("...fail")
            return 0, pointsPossible
            
         print("...pass")
         return pointsPossible, pointsPossible

         
   except IOError:
      print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
      traceback.print_exc(file=sys.stdout) 
      return 0, 0
   

def LTM(testFileName):
  
   print("Running link transmission model test: " + str(testFileName) + "...", end='')
   inflow = list()
   targetSendingFlow = list()
   targetReceivingFlow = list()
   targetUpstreamCount = list()
   targetDownstreamCount = list()
   
   try:
      with open(testFileName, "r") as testFile:
         # Read test information
         try:
            fileLines = testFile.read().splitlines()
            pointsPossible = IS_MISSING
            timeHorizon = IS_MISSING
            for line in fileLines:
               # Ignore comments and blank lines
               if len(line.strip()) == 0 or line[0] == '#':
                  continue
                   
               # Set points possible
               if pointsPossible == IS_MISSING:
                  pointsPossible = int(line)
                  continue
                  
               # Set problem parameters
               if timeHorizon == IS_MISSING:
                  inputs = line.split(",")
                  timeHorizon = int(inputs[0])
                  capacity = int(inputs[1])
                  freeFlowTime = int(inputs[2])
                  backwardWaveTime = int(inputs[3])
                  jamDensity = float(inputs[4])
                  if len(inputs) > 5:
                     firstOutflow = int(inputs[5])
                  else:
                     firstOutflow = 0
                  continue
                  
               # Read problem data
               timestepData = line.split(",")
               inflow.append(float(timestepData[0]))
               targetUpstreamCount.append(float(timestepData[1]))
               targetDownstreamCount.append(float(timestepData[2]))
               targetReceivingFlow.append(float(timestepData[3]))
               targetSendingFlow.append(float(timestepData[4]))               
                  
         except:
            print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
            traceback.print_exc(file=sys.stdout)
            return 0, 0
            
         # Now run the actual test
         try:
            testLink = linkModel.LinkTransmissionModelLink(1 * HOURS, 1 / freeFlowTime, 1 / backwardWaveTime, jamDensity, 1 * MILES, capacity)
            enteringFlow = dict()
            leavingFlow = dict()
            for t in range(timeHorizon):
               upstreamCount, downstreamCount = testLink.upstreamCount(t), testLink.downstreamCount(t)
               sendingFlow, receivingFlow = testLink.linkUpdate(t)

               enteringFlow['link'] = min(inflow[t], receivingFlow)
               leavingFlow['link'] = sendingFlow if t >= firstOutflow else 0
              
               testLink.flowIn(enteringFlow)
               testLink.flowOut(leavingFlow)
              
               correct = True
               correct = correct and check("sending flow", t, sendingFlow, targetSendingFlow[t], 0.01)
               correct = correct and check("receiving flow", t, receivingFlow, targetReceivingFlow[t], 0.01)            
               correct = correct and check("upstream count", t, upstreamCount, targetUpstreamCount[t], 0.01)            
               correct = correct and check("downstream count", t, downstreamCount, targetDownstreamCount[t], 0.01)            
               if correct == False:               
                  print("\n...fail")
                  return 0, pointsPossible
         except utils.NotYetAttemptedException:
            print("...not yet attempted")
            return 0, pointsPossible
         except:
            print("\nException raised, attempting to continue:")
            traceback.print_exc(file=sys.stdout)                     
            print("...fail")
            return 0, pointsPossible
            
         print("...pass")
         return pointsPossible, pointsPossible

         
   except IOError:
      print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
      traceback.print_exc(file=sys.stdout) 
      return 0, 0
   
def CTM(testFileName):
  
   print("Running cell transmission model test: " + str(testFileName) + "...", end='')
   inflow = list()
   targetSendingFlow = list()
   targetReceivingFlow = list()
   targetUpstreamCount = list()
   targetDownstreamCount = list()
   
   try:
      with open(testFileName, "r") as testFile:
         # Read test information
         try:
            fileLines = testFile.read().splitlines()
            pointsPossible = IS_MISSING
            timeHorizon = IS_MISSING
            for line in fileLines:
               # Ignore comments and blank lines
               if len(line.strip()) == 0 or line[0] == '#':
                  continue
                   
               # Set points possible
               if pointsPossible == IS_MISSING:
                  pointsPossible = int(line)
                  continue
                  
               # Set problem parameters
               if timeHorizon == IS_MISSING:
                  inputs = line.split(",")
                  timeHorizon = int(inputs[0])
                  timestep = int(inputs[1])
                  capacity = int(inputs[2])
                  freeFlowSpeed = int(inputs[3])
                  backwardWaveSpeed = int(inputs[4])
                  jamDensity = float(inputs[5])
                  length = float(inputs[6])
                  if len(inputs) > 7:
                     firstOutflow = int(inputs[7])
                  else:
                     firstOutflow = 0
                  continue
                  
               # Read problem data
               timestepData = line.split(",")
               inflow.append(float(timestepData[0]))
               targetUpstreamCount.append(float(timestepData[1]))
               targetDownstreamCount.append(float(timestepData[2]))
               targetReceivingFlow.append(float(timestepData[3]))
               targetSendingFlow.append(float(timestepData[4]))               
                  
         except:
            print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
            traceback.print_exc(file=sys.stdout)
            return 0, 0
            
         # Now run the actual test
         try:
            testLink = linkModel.CellTransmissionModelLink(timestep, freeFlowSpeed, backwardWaveSpeed, jamDensity, length, capacity)
            enteringFlow = dict()
            leavingFlow = dict()
            for t in range(timeHorizon):
               upstreamCount, downstreamCount = testLink.upstreamCount(t), testLink.downstreamCount(t)
               sendingFlow, receivingFlow = testLink.linkUpdate(t)

               enteringFlow['link'] = min(inflow[t], receivingFlow)
               leavingFlow['link'] = sendingFlow if t >= firstOutflow else 0
              
               testLink.flowIn(enteringFlow)
               testLink.flowOut(leavingFlow)
              
               correct = True
               correct = correct and check("sending flow", t, sendingFlow, targetSendingFlow[t], 0.1)
               correct = correct and check("receiving flow", t, receivingFlow, targetReceivingFlow[t], 0.1)            
               correct = correct and check("upstream count", t, upstreamCount, targetUpstreamCount[t], 0.1)            
               correct = correct and check("downstream count", t, downstreamCount, targetDownstreamCount[t], 0.1)            
               if correct == False:               
                  print("...fail")
                  return 0, pointsPossible
         except utils.NotYetAttemptedException:
            print("...not yet attempted")
            return 0, pointsPossible
         except:
            print("\nException raised, attempting to continue:")
            traceback.print_exc(file=sys.stdout)                     
            print("\n...fail")
            return 0, pointsPossible
            
         print("...pass")
         return pointsPossible, pointsPossible

         
   except IOError:
      print("\nError running test %s, attempting to continue with remaining tests.  Exception details: " % testFileName)
      traceback.print_exc(file=sys.stdout) 
      return 0, 0
      